import os
from django.conf import settings
from django.utils.module_loading import import_string
# from rest_framework.settings import APISettings

CHANNELS_BINDING = getattr(settings, 'CHANNELS_BINDING', {})
DEFAULT_CHANNELS_BINDING = {
    "AUTHENTIFICATION_CLASSES": (
        'channels_binding.authentification.DefaultDjangoUser',
    ),
    "DEFAULT_PAGE_SIZE": 25,
    "ANONYMOUS_CONNECTION_ALLOWED": False,
}


def get_default(name):
    return CHANNELS_BINDING.get(name, DEFAULT_CHANNELS_BINDING[name])


AUTHENTIFICATION_CLASSES = []
classes = get_default("AUTHENTIFICATION_CLASSES")
if not isinstance(classes, (list, tuple)):
    classes = [classes]
for string in classes:
    try:
        AUTHENTIFICATION_CLASSES.append(import_string(string))
    except ImportError as e:
        msg = "Could not import '%s' for Channels Bindings setting '%s': %s." % (string, e.__class__.__name__, e)
        raise ImportError(msg)

DEFAULT_PAGE_SIZE = get_default("DEFAULT_PAGE_SIZE")
ANONYMOUS_CONNECTION_ALLOWED = bool(get_default("ANONYMOUS_CONNECTION_ALLOWED"))



from channels.db import database_sync_to_async
from django.forms import modelform_factory
from ..utils import (
    bind, db_sync, sync
)

from . import settings as self_settings

__all__ = [
    'AsyncSearchModelBinding',
    'AsyncRetrieveModelBinding',
    'AsyncSaveModelBinding',
]


class AsyncSearchModelBinding(object):

    @database_sync_to_async
    def get_search_data(self, data):
        queryset = self.get_queryset(data)
        queryset = self.filter_queryset(queryset, data)
        queryset = self.paginate(queryset, data)
        search_data = self.serialize_queryset(queryset, data)
        return search_data

    @bind('search')
    async def search(self, data, *args, **kwargs):
        await self.reflect(f'search', await self.get_search_data(data), *args, **kwargs)


class AsyncRetrieveModelBinding(object):

    @database_sync_to_async
    def get_retrieve_data(self, data):
        instance = self.get_object(data)
        retrieve_data = self.serialize(instance, data)
        retrieve_data.update(id=instance.pk)
        return retrieve_data

    @bind('retrieve')
    async def retrieve(self, data, *args, **kwargs):
        await self.reflect('retrieve', await self.get_retrieve_data(data), *args, **kwargs)


class AsyncSaveModelBinding(object):

    @database_sync_to_async
    def get_save_data_form(self, data):
        form = self.get_form(data)
        if self.form_is_valid(form, data):
            self.save_form(form, data)
        save_data = self.serialize_form(form, data)
        return save_data, form

    def form_is_valid(self, form, data):
        return form.is_valid()

    def get_form_fields(self, data):
        return self.fields

    def save_form(self, form, data):
        return form.save()

    def get_form(self, data):
        instance = self.get_object(data)
        fields = self.get_form_fields(data)
        for name in fields:
            if name not in data:
                data[name] = getattr(instance, name, None)
        if self.model and not self.model_form:
            form = modelform_factory(self.model, fields=fields)(data, instance=instance)
        else:
            form = model_form(data, instance=instance)
        return form

    @bind('save')
    async def save(self, data, *args, **kwargs):
        save_data, form = await self.get_save_data_form(data)
        await self.reflect('save', save_data, *args, **kwargs)
        if not form.errors:
            await self.dispatch('retrieve', await self.get_retrieve_data(data), *args, **kwargs)



import json
import os
import requests
import traceback
import logging
from channels.exceptions import DenyConnection
from channels.generic.websocket import AsyncWebsocketConsumer
from channels.db import database_sync_to_async
from django.conf import settings
from .bindings.registry import (
    registered_binding_classes,
    registered_binding_events
)
from .utils import bind, JSONEncoder
from . import settings as self_settings


logger = logging.getLogger(__name__)
__all__ = [
    'AsyncConsumer',
]


class AsyncConsumer(AsyncWebsocketConsumer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None
        self.user_id = None
        self.user_group_name = None
        self.groups = set('__all__')
        self.actions = {}
        self.authentifications = [cls() for cls in self_settings.AUTHENTIFICATION_CLASSES]

    @database_sync_to_async
    def get_user(self):
        self.user = None
        self.user_id = None
        self.user_group_name = None
        for auth in self.authentifications:
            if hasattr(auth, 'get_user'):
                self.user = auth.get_user(self)
                if self.user:
                    self.user_id = getattr(self.user, 'pk', getattr(self.user, 'id', getattr(self.user, 'key')))
                    if self.user_id:
                        self.user_group_name = f'user.{self.user_id}'
                    break

    async def connect(self):
        try:
            self.user = self.get_user()
            if self.user:
                self.bindings_by_class = {bc: bc(self) for bc in registered_binding_classes}
                await self.subscribe('__all__')
                if self.user_group_name:
                    await self.subscribe(user_group_name)
                await self.accept()
                # await self.send('auth.roles', self.user.roles)
            else:
                if self_settings.ANONYMOUS_CONNECTION_ALLOWED:
                    await self.accept()
                else:
                    raise DenyConnection("Invalid Token")
                    await self.close()
        except Exception as e:
            logger.error(traceback.format_exc())
            raise DenyConnection(traceback.format_exc())
            await self.close()

    # On client disconnect
    async def disconnect(self, close_code):
        for group in set(self.groups):
            await self.unsubscribe(group)

    async def subscribe(self, group):
        self.groups.add(group)
        await self.channel_layer.group_add(group, self.channel_name)

    async def unsubscribe(self, group):
        self.groups.remove(group)
        await self.channel_layer.group_discard(group, self.channel_name)

    # On client send a message, pass throught the corresponding action_ (event)

    async def receive(self, text_data):
        try:
            data = json.loads(text_data)
            event_hash = data['event'].split('#', 1)
            event = event_hash[0]
            hash = event_hash[-1] if len(event_hash) == 2 else None
            payload = data.get('data', {})
            events = registered_binding_events.get(event, [])
            counter = 0
            for binding_class, method_name in events:
                binding = self.bindings_by_class.get(binding_class, None)
                if binding:
                    await self.subscribe(binding.stream)  # TODO: auto unsubscribe or get subscribe from front
                    if not isinstance(payload, dict):
                        payload = {}
                    await getattr(binding, method_name)(payload, hash=hash)
                    counter += 1
            if not counter:
                await self.send('error', f'No binding found for {event}#{hash}')
        except Exception as e:
            logger.error(traceback.format_exc())
            await self.send('error', traceback.format_exc())

    # Send a event message
    async def send(self, event, data, group=None, user=None, hash=None):

        if hash:
            event = f'{event}#{hash}'
        message = await self.encode({'event': event, 'data': data})
        # Private
        if user:
            # Self to a specific user
            await self.channel_layer.group_send(
                f'user.{user.id}',
                {'type': 'channel_message', 'message': message}
            )

        # Dispatch or group == __all__ to broadcast
        elif group:
            # Self to a specific group
            await self.channel_layer.group_send(
                group,
                {'type': 'channel_message', 'message': message}
            )

        # Reflect
        else:
            # Self to current thread ~= the connected user
            await super().send(text_data=message)

    async def encode(self, message):
        return json.dumps(message, cls=JSONEncoder)

    # Receive message from the group
    async def channel_message(self, message):
        await super().send(text_data=message['message'])



import json
import decimal
import datetime
from channels.db import database_sync_to_async
from asgiref.sync import async_to_sync
try:
    from psycopg2.extras import NumericRange
except ImportError as e:
    NumericRange = None

__all__ = [
    'bind',
    'db_sync',
    'sync',
    'JSONEncoder',
]


def bind(*args, **kwargs):
    """
    Used to mark a method on a Binding that should be binded for event receive.
    """
    def decorator(func):
        func.is_bind = True
        func.args = args
        func.kwargs = kwargs
        return func
    return decorator


class DbSync():
    """
    Usage : with db_sync:
    """

    @database_sync_to_async
    def __enter__(self):
        return {}

    @database_sync_to_async
    def __exit__(self, type, value, traceback):
        # Exception handling here
        pass


db_sync = DbSync()


class Sync():
    """
    Usage : with sync:
    """

    @async_to_sync
    def __enter__(self):
        return None

    @async_to_sync
    def __exit__(self, type, value, traceback):
        # Exception handling here
        pass


sync = Sync()


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, '__json__'):
            return obj.__json__()
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, bytes):
            return obj.decode('utf-8')
        elif isinstance(obj, memoryview):
            return obj.tobytes().decode('utf-8')
        elif NumericRange and isinstance(obj, NumericRange):
            return [obj.lower, obj.upper]
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        return super().default(obj)



from channels.db import database_sync_to_async

__all__ = [
    'AsyncSerializerBinding',
]


class AsyncSerializerBinding(object):

    serializer_class = None

    def serialize(self, instance, *args, **kwargs):
        # kwargs['context'] = self.get_serializer_context()
        if not self.serializer_class:
            if hasattr(instance, '__json__'):
                return getattr(instance, '__json__')(**kwargs)
            elif hasattr(instance, 'to_json'):
                return getattr(instance, 'to_json')(**kwargs)
            elif hasattr(instance, 'json'):
                return getattr(instance, 'json')
            else:
                return {}
        else:
            return self.serializer_class(instance, *args, **kwargs).data

    def serialize_queryset(self, queryset, *args, **kwargs):
        if not self.serializer_class:
            rows = [self.serialize(inst, *args, **kwargs) for inst in queryset]
            return dict(
                page=queryset.number,
                limit=queryset.paginator.per_page,
                count=queryset.paginator.count,
                rows=rows
            )
        else:
            return self.serializer_class(queryset, *args, **kwargs).data

    def serialize_form(self, form, *args, **kwargs):
        if form.errors:
            return {'errors': form.errors}
        else:
            return {'success': True}



from django.core.exceptions import ObjectDoesNotExist
from django.core.paginator import Paginator, Page
from django.forms import modelform_factory
from channels.db import database_sync_to_async

__all__ = [
    'AsyncModelBinding',
]


class AsyncModelBinding(object):

    queryset = None
    # mark as abstract
    model = None
    lookup_field = 'pk'
    data_pk = 'id'
    page_size = 25
    model_form = None
    fields = []  # hack to pass cls.register() without ValueError

    def get_queryset(self, data):
        if not self.queryset:
            if self.model:
                return self.model.objects.all()
        else:
            return self.queryset.all()

    def filter_in(self, queryset, name, data, lookup=None):
        if name in data:
            value = data[name]
            if isinstance(value, list):
                if len(value):
                    return queryset.filter(**{f'{lookup or name}__in': value})
                else:
                    return queryset
            else:
                return queryset.filter(**{f'{lookup or name}': value})
        else:
            return queryset

    def filter_queryset(self, queryset, data):
        return queryset

    def paginate(self, queryset, data, limit=None):
        limit = limit or self.page_size or self_settings.DEFAULT_PAGE_SIZE
        try:
            page = int(data['page'])
        except (KeyError, ValueError, TypeError) as e:
            page = 1
        paginator = Paginator(queryset, max(10, min(100, int(limit))))
        return paginator.get_page(page)

    def get_object(self, data):
        print(data)
        pk = data.get(self.data_pk, None)
        try:
            return self.model.objects.get(pk=pk)
        except self.model.DoesNotExist:
            raise Exception(f'{self.stream} pk:{pk} Does Not Exist')



import glob
import os
import importlib.util
from django.apps import AppConfig
from django.utils.module_loading import autodiscover_modules, module_has_submodule
from importlib import import_module
from importlib.util import find_spec as importlib_find


class AppConfig(AppConfig):
    name = 'channels_binding'
    verbose_name = "Channels Binding"

    def ready(self):

        from django.apps import apps
        module_to_search = 'bindings'
        autodiscover_modules(module_to_search)
        # autodiscover bindings
        for app_config in apps.get_app_configs():
            try:
                module = import_module('%s.%s' % (app_config.name, module_to_search))
                try:
                    package_name = module.__name__
                    package_path = module.__path__[0]
                    for file in glob.glob(os.path.join(package_path, '*.py')):
                        filename = os.path.basename(file)
                        try:
                            module = import_module('%s.%s.%s' % (app_config.name, module_to_search, filename))
                        except Exception as e:
                            pass
                except AttributeError as e:
                    print(e)
            except ModuleNotFoundError as e:
                pass
            except Exception as e:
                print(e)

        # for base_start, pattern, base_stop in [
        #     ['app.bindings', os.path.join(os.path.dirname(__file__), '../bindings/*.py'), ''],
        #     ['apps', os.path.join(os.path.dirname(__file__), '../apps/*/bindings.py'), 'bindings']
        # ]:
        #     # Dynamicaly loads of all consumer bindings
        #     for file in glob.glob(pattern):
        #         file = os.path.abspath(file)
        #         spec = importlib.util.spec_from_file_location('.'.join([o for o in [base_start, os.path.basename(file), base_stop] if o]), file)
        #         module = importlib.util.module_from_spec(spec)
        #         spec.loader.exec_module(module)



import traceback
import os
from channels.db import database_sync_to_async
from django.conf import settings
from .registry import (
    RegisteredBindingMetaClass,
)
from .models import (
    AsyncModelBinding,
)
from .serializers import (
    AsyncSerializerBinding,
)
from .events import (
    AsyncSearchModelBinding,
    AsyncRetrieveModelBinding,
    AsyncSaveModelBinding,
)
from ..utils import (
    bind
)

__all__ = [
    'AsyncBinding',
    'bind'
]


class AsyncBindingBase(metaclass=RegisteredBindingMetaClass):

    stream = None
    permission_classes = ()

    def __init__(self, consumer):
        self.consumer = consumer
        self.user = consumer.user

    # Respond to the current socket
    async def send(self, event, *args, **kwargs):
        stream = kwargs.get('stream', self.stream) or self.stream
        await self.consumer.send(f'{stream}.{event}', *args, **kwargs)

    # Respond to the current socket
    async def reflect(self, *args, **kwargs):
        await self.send(*args, **kwargs)

    # Respond to the current streamed group attached sockets

    async def dispatch(self, *args, **kwargs):
        await self.send(*args, group=kwargs.get('stream', self.stream), **kwargs)

    # Respond to all sockets
    async def broadcast(self, *args, **kwargs):
        await self.send(*args, group='__all__', **kwargs)

    # Respond to the current streamed group attached sockets
    async def subscribe(self, group=None):
        await self.consumer.subscribe(group or self.stream)

    # Respond to the current streamed group attached sockets
    async def unsubscribe(self, group=None):
        await self.consumer.unsubscribe(group or self.stream)


class AsyncBinding(
    AsyncBindingBase,
    AsyncModelBinding,
    AsyncSerializerBinding,
    AsyncSearchModelBinding,
    AsyncRetrieveModelBinding,
    AsyncSaveModelBinding,
):
    pass



__all__ = [
    'RegisteredBindingMetaClass',
    'registered_binding_classes',
    'registered_binding_events',
]


registered_binding_classes = set()
registered_binding_events = dict()


class RegisteredBindingMetaClass(type):
    def __new__(cls, clsname, superclasses, attributedict):
        binding_class = type.__new__(cls, clsname, superclasses, attributedict)
        if superclasses:
            stream = None
            if binding_class.stream:
                stream = binding_class.stream
            elif binding_class.model:
                stream = f'{binding_class.model._meta.app_label}.{binding_class.model._meta.object_name}'
            binding_class.stream = stream
            binding_class._registred_actions = {}

            for method_name in dir(binding_class):
                method = getattr(binding_class, method_name)
                is_bind = getattr(method, 'is_bind', False)
                if is_bind:
                    args = getattr(method, 'args', [])
                    kwargs = getattr(method, 'kwargs', {})
                    name = kwargs.get('name', args[0] if len(args) else method_name) or method_name
                    event = f'{stream}.{name}'
                    events = registered_binding_events.setdefault(event, [])
                    events.append([binding_class, method_name])

            if stream:
                registered_binding_classes.add(binding_class)
                print(f'\033[40;92m\033[1m[Bindings registered]\033[0m \033[40;92m{stream} ({binding_class})\033[0m')
        return binding_class



import os
import sys
import setuptools

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (2, 7)
EXCLUDE_FROM_PACKAGES = [
    'tests', 'build', 'dist',
]
if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write("""
==========================
Unsupported Python version
==========================


This version of Django requires Python {}.{}, but you're trying to
install it on Python {}.{}.

""".format(*(REQUIRED_PYTHON + CURRENT_PYTHON)))
    sys.exit(1)


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


version = __import__('channels_binding').__version__

setuptools.setup(
    name='channels_binding',
    version=version,
    # python_requires='>={}.{}'.format(*REQUIRED_PYTHON),
    author="Damien Autrusseau",
    author_email="damien.autrusseau@gmail.com",
    description="Channels API Bindings system",
    license="MIT",
    keywords="channels api binding websocket rest django",
    url="https://pypi.org/project/channels-binding",
    # packages=['tests'],
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=EXCLUDE_FROM_PACKAGES),
    include_package_data=True,
    install_requires=[
        'Django>=1.8',
        'channels>=2.1.3',
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



# from rest_framework import routers
from django.conf import settings
from django.http import HttpResponse
from django.urls import include, path, re_path
from django.conf.urls.static import static
from channels.routing import ProtocolTypeRouter, URLRouter, ChannelNameRouter
from channels.sessions import SessionMiddlewareStack
from .consumers import AsyncConsumer


async_root_application = ProtocolTypeRouter({
    # (http->django views is added by default)
    'websocket': SessionMiddlewareStack(
        URLRouter([
            path('', AsyncConsumer, name="root"),
        ])
    )
})


# router = routers.DefaultRouter()
# router.registry.extend(test_router.registry)

# urlpatterns = [
#     # re_path(r"^$", lambda request: HttpResponse(f"[OK] WSAPI CLEARVERSION")
#     # ),
#     path(r'', include((router.urls, 'api'), namespace='api')),
#     path(r'api-auth/', include('rest_framework.urls')),
#     path(r'graphmodels/', graphmodels)
# ]
